
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "filter.h"


/* adapted from The Csound moogvcf opcode in Opcodes/biquad.c
 * Implementation of:
 * http://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.68.7084&rep=rep1&type=pdf
 */

#ifndef M_E
#define M_E 2.71828182845904523536
#endif

void *filter_construct(int sr)
{
    FILTER_DATA *data = (FILTER_DATA *) malloc(sizeof(FILTER_DATA));
    data->in = (double *) malloc(sizeof(double));

    data->xnm1 = data->y1nm1 = data->y2nm1 = data->y3nm1 = 0.0;
    data->y1n  = data->y2n   = data->y3n   = data->y4n   = 0.0;

    data->sr = sr;

    return (void *) data;
}

void filter_process(void *d)
{
    FILTER_DATA *data = (FILTER_DATA *) d;
//    double y1, y2, y3;
    double x;
    double xnm1 = data->xnm1, y1nm1 = data->y1nm1, y2nm1 = data->y2nm1, y3nm1 = data->y3nm1;
    double y1n  = data->y1n, y2n = data->y2n, y3n = data->y3n, y4n = data->y4n;

    x = *data->in - data->k*data->y4n;
    y1n   = (x  + xnm1 ) * data->pp1d2 - data->kp * y1n;
    y2n   = (y1n + y1nm1) * data->pp1d2 - data->kp * y2n;
    y3n   = (y2n + y2nm1) * data->pp1d2 - data->kp * y3n;
    y4n   = (y3n + y3nm1) * data->pp1d2 - data->kp * y4n;

    /* Clipper band limited sigmoid */
    y4n   = y4n - y4n * y4n * y4n / 6.0;
    xnm1  = x;       /* Update Xn-1  */
    y1nm1 = y1n;      /* Update Y1n-1 */
    y2nm1 = y2n;      /* Update Y2n-1 */
    y3nm1 = y3n;      /* Update Y3n-1 */
    x   = y4n;

    data->xnm1 = xnm1; data->y1nm1 = y1nm1; data->y2nm1 = y2nm1; data->y3nm1 = y3nm1;
    data->y1n  = y1n;  data->y2n  = y2n; data->y3n = y3n; data->y4n = y4n;

    *data->out = x;
}

void filter_destroy(void *data)
{
    FILTER_DATA *filterdata = (FILTER_DATA *) data;
    free(filterdata->in);
}

void filter_update(FILTER_DATA *data)
{
    double f = 2 * data->cutoff / data->sr;
    data->kp = 3.6*f - 1.6*f*f -1; /* (Empirical tunning) */
    data->pp1d2 = (data->kp+1)*0.5;
    double scale = pow(M_E, ((1-data->pp1d2)*1.386249));
    data->k = data->res*scale;
}


void filter_set_param(void *d, int index, double value)
{
    FILTER_DATA *data = (FILTER_DATA *) d;
    switch(index) {
    case FILTER_CF:
        data->cutoff = value;
        filter_update(data);
        break;
    case FILTER_RES:
        data->res = value;
        filter_update(data);
        break;
    }
    return;
}


double *filter_get_input(void *data, int index)
{
    FILTER_DATA *d = (FILTER_DATA *) data;
    switch (index) {
    case FILTER_IN:
        return d->in;
    case FILTER_CF:
        return &d->cutoff;
    case FILTER_RES:
        return &d->res;
    }

    return NULL;
}


void filter_set_output(void *data, int index, double *out_buffer)
{
    FILTER_DATA *d = (FILTER_DATA *) data;
    switch (index) {
    case 0:
        d->out = out_buffer;
        break;
    }
}


void filter_get_ugen_info(ugen_t *ugen_data)
{
    strcpy(ugen_data->name, "filter");
    ugen_data->process = filter_process;
    ugen_data->destroy = filter_destroy;
    ugen_data->set_param = filter_set_param;
    ugen_data->get_input = filter_get_input;
    ugen_data->set_output = filter_set_output;
}
